@@ -40,9 +40,9 @@ def etree_to_dict(elem):
                 defd[key].append(val)
         result = {}
         for key, val in defd.items():
-            print(f"{key}, {val[0]}, {type(val[0])}, {has_size}")
-            is_array = len(val) > 1 or ( has_size and type(val[0]) is dict )
-            if not elem.tag in result:
+            #             print(f"{key}, {val[0]}, {type(val[0])}, {has_size}")
+            is_array = len(val) > 1 or (has_size and type(val[0]) is dict)
+            if elem.tag not in result:
                 result[elem.tag] = {}
             if is_array:
                 result[elem.tag][key] = val
