@@ -20,18 +20,19 @@ __email__ = "reynolds@biology.utah.edu, sam.forester@utah.edu, tonyw@honestpuck.
 __copyright__ = "Copyright (c) 2021 University of Utah, School of Biological Sciences and Copyright (c) 2020 Tony Williams"
 __license__ = "MIT"
 __date__ = "2020-09-21"
-__version__ = "0.4.4"
+__version__ = "0.4.6"
 
 
-# pylint: disable=relative-beyond-top-level
-from .api import API
-from pprint import pprint
 import copy
 import json
 import logging
 import os.path
 import re
-import sys
+from pprint import pprint
+from sys import stderr
+
+# pylint: disable=relative-beyond-top-level
+from .api import API
 
 __all__ = (
     "AdvancedComputerSearches",
@@ -97,8 +98,8 @@ def valid_records():
     valid = tuple(
         x
         for x in __all__
-        if not x
-        in [
+        if x
+        not in [
             # Exclude list, add all non-Jamf Record classes here
             "JamfError"
         ]
@@ -154,8 +155,8 @@ class ClassicSwagger(metaclass=Singleton):
             "DirectoryBindings": post_template2,
             "Ebooks": post_template2,
             "JSONWebTokenConfigurations": post_template2,
-            #"LicensedSoftware": post_template2,
-            #"LDAPServers": post_template2,
+            # "LicensedSoftware": post_template2,
+            # "LDAPServers": post_template2,
             "MacApplications": post_template2,
             "ManagedPreferenceProfiles": post_template2,
             "MobileDevices": post_template2,
@@ -204,7 +205,7 @@ class ClassicSwagger(metaclass=Singleton):
                 "general": {"name": "%NAME%", "target_version": "%VERSION%"}
             },
             "PatchSoftwareTitles": {"name_id": "%NAME%", "source_id": "1"},
-            #"RestrictedSoftware": {"general":{"name":"%NAME%","process_name":"%NAME%"}},
+            # "RestrictedSoftware": {"general":{"name":"%NAME%","process_name":"%NAME%"}},
             "UserGroups": {"general": {"name": "%NAME%"}, "is_smart": False},
             "WebHooks": {
                 "event": "ComputerAdded",
@@ -358,21 +359,16 @@ class Record:
     Just in case that's not confusing enough the id tag is not always 'id'.
     """
 
-    def __new__(cls, *a, **kw):
+    def __new__(cls, *args, **kw):
         """
         returns existing record if one has been instantiated
         """
-        jamf_id = int(a[0])
-        if type(a[1]) is str:
-            name = a[1]
-        else:
-            name = a[1][0]
+        jamf_id = int(args[0])
         if not hasattr(cls, "_instances"):
             cls._instances = {}
         swag = ClassicSwagger()
         plural = eval(cls.plural_class)
         api = API()
-        _data = {}
         if jamf_id == 0:
             if not swag.is_action_valid(plural, "post"):
                 print(
@@ -382,21 +378,44 @@ class Record:
                 )
                 return None
             s1, s2, end = swag.swagger(plural, "s1, s2, end")
-            if cls.plural_class == "PatchPolicies":
-                if len(a[1]) < 3:
-                    raise JamfError(f"patchpolicies requires 3 args to create records")
-                softwaretitleconfigid = a[1][1]
-                end = f"patchpolicies/softwaretitleconfig/id/{softwaretitleconfigid}"
-                out = {s1: swag.post_template(cls.plural_class, name)}
-                t = out["patch_policy"]["general"]["target_version"]
-                out["patch_policy"]["general"]["target_version"] = t.replace(
-                    "%VERSION%", a[1][2]
-                )
+            use_template = True
+            if type(args[1]) is list:
+                if type(args[1][0]) is str:
+                    name = args[1][0]
+                elif type(args[1][0]) is dict:
+                    name = ""
+                    use_template = False
+            elif type(args[1]) is str or args[1] is None:
+                name = args[1]
+            if use_template:
+                if cls.plural_class == "PatchPolicies":
+                    if len(args[1]) < 3:
+                        raise JamfError(
+                            "patchpolicies requires 3 args to create records"
+                        )
+                    softwaretitleconfigid = args[1][1]
+                    end = (
+                        f"patchpolicies/softwaretitleconfig/id/{softwaretitleconfigid}"
+                    )
+                    out = {s1: swag.post_template(cls.plural_class, name)}
+                    t = out["patch_policy"]["general"]["target_version"]
+                    out["patch_policy"]["general"]["target_version"] = t.replace(
+                        "%VERSION%", args[1][2]
+                    )
+                else:
+                    end = f"{end}0"
+                    out = {s1: swag.post_template(cls.plural_class, name)}
+                _data = api.post(end, out)
             else:
                 end = f"{end}0"
-                out = {s1: swag.post_template(cls.plural_class, name)}
-            _data = api.post(end, out)
+                _data = api.post(end, args[1][0])
             jamf_id = int(_data[s2]["id"])
+        else:
+            _data = {}
+            try:
+                name = args[1]
+            except NameError:
+                name = ""
         if jamf_id not in cls._instances:
             rec = super(Record, cls).__new__(cls)
             rec.cls = cls
@@ -433,10 +452,10 @@ class Record:
         return False
 
     def __str__(self):
-        if self.name:
+        if self.name is not None:
             return self.name
         else:
-            return "Empty name"
+            return ""
 
     def __repr__(self):
         return f"{self.__class__.__name__}({self.id}, {self.name!r})"
@@ -447,7 +466,7 @@ class Record:
         if not self.s.is_action_valid(self.plural, "get"):
             raise JamfError(f"get({end}) is an invalid action for get")
         results = self.api.get(end)
-        if not s1 in results:
+        if s1 not in results:
             print("---------------------------------------------\nData dump\n")
             pprint(results)
             raise JamfError(f"Endpoint {end} has no member named {s1} (s1).")
@@ -477,48 +496,41 @@ class Record:
             self.refresh()
         return self._data
 
-    def get_path2(self, path, placeholder, index=0):
+    def get_path_worker(self, path, placeholder, index=0):
         current = path[index]
         if type(placeholder) is dict:
             if current in placeholder:
                 if index + 1 >= len(path):
                     return placeholder[current]
-                placeholder = self.get_path2(path, placeholder[current], index + 1)
+                placeholder = self.get_path_worker(
+                    path, placeholder[current], index + 1
+                )
             else:
-                return None
+                raise NotFound
         elif type(placeholder) is list:
             # I'm not sure this is the best way to handle arrays...
             result = []
             for item in placeholder:
                 if current in item:
                     if index + 1 < len(path):
-                        result.append(self.get_path2(path, item[current], index + 1))
+                        result.append(
+                            self.get_path_worker(path, item[current], index + 1)
+                        )
                     else:
                         result.append(item[current])
             placeholder = result
-        elif placeholder == None:
-            return None
-        else:
-            print("Something went wrong in get_path2")
-            exit()
         return placeholder
 
     def get_path(self, path):
         if not self._data:
             self.refresh()
-        try:
-            result = self.get_path2(path.split("/"), self._data)
-        except NotFound as error:
-            print("Not Found")
-            result = []
-        if type(result) is list or type(result) is dict or result == None:
-            return result
-        return [result]
+        result = self.get_path_worker(path.split("/"), self._data)
+        return result
 
     def force_array(self, parent, child_name):
         _data = parent[child_name]
         if "size" in parent:
-            if int(parent["size"]) > 1:
+            if int(parent["size"]) > 0:
                 return _data
         else:
             if type(_data) is list:
@@ -544,7 +556,7 @@ class Record:
                 pprint(placeholder)
                 return False
         else:
-            print(f"Error: empty data:")
+            print("Error: empty data:")
             pprint(placeholder)
             return False
 
@@ -618,12 +630,25 @@ class Records:
     def recordWithId(self, x):
         if not self.data:
             self.refresh()
+        if type(x) is str:
+            x = int(x)
         return self._jamf_ids.get(x)
 
     def recordWithName(self, x):
+        stderr(
+            "WARNING: recordWithName deprecated, use recordsWithName (it turns out names are not unique)."
+        )
+        names = self.recordsWithName(self, x)[0]
+        if len(names) > 1:
+            stderr(
+                "There is more than one record with the name you are searching for! Only the first one is being used."
+            )
+        return names[0]
+
+    def recordsWithName(self, x):
         if not self.data:
             self.refresh()
-        return self._names.get(x)
+        return [self._names.get(x)]
 
     def recordsWithRegex(self, x):
         if not self.data:
@@ -639,15 +664,12 @@ class Records:
         lst = self.api.get(p1)  # e.g. categories
         if p2 in lst:
             self.data = lst[p2]
-            if not self.data or not "size" in self.data or self.data["size"] == "0":
+            if not self.data or "size" not in self.data or self.data["size"] == "0":
                 self._records = {}
                 self._names = {}
                 self._jamf_ids = {}
             elif p3 in self.data:  # e.g. category
-                if self.data["size"] == "1":
-                    records = [self.data[p3]]
-                else:
-                    records = self.data[p3]
+                records = self.data[p3]
                 for d in records:
                     c = self.singular_class(d[id1], d["name"])  # e.g. id1 = "id"
                     c.plural_class = self.cls
@@ -772,14 +794,20 @@ class Computer(Record):
         if not hasattr(plural_cls, "computers"):
             plural_cls.computers = {}
         plural_cls.computers[self.name] = True
-        apps = self.get_path("software/applications/application/path")
-        versions = self.get_path("software/applications/application/version")
+        try:
+            apps = self.get_path("software/applications/application/path")
+        except NotFound:
+            apps = None
+        try:
+            versions = self.get_path("software/applications/application/version")
+        except NotFound:
+            versions = None
         if apps:
             for ii, app in enumerate(apps):
                 ver = versions[ii]
-                if not app in plural_cls.app_list:
+                if app not in plural_cls.app_list:
                     plural_cls.app_list[app] = {}
-                if not ver in plural_cls.app_list[app]:
+                if ver not in plural_cls.app_list[app]:
                     plural_cls.app_list[app][ver] = {}
                 plural_cls.app_list[app][ver][self.name] = True
 
@@ -787,7 +815,9 @@ class Computer(Record):
 class Computers(Records, metaclass=Singleton):
     # http://localhost/computers.html?queryType=COMPUTERS&query=
     singular_class = Computer
-    sub_commands = {"apps": {"required_args": 0, "args_description": ""}}
+    sub_commands = {
+        "apps": {"required_args": 0, "args_description": ""},
+    }
 
     def apps_print_after(self):
         print("application,version,", end="")
@@ -1023,11 +1053,11 @@ class OSXConfigurationProfiles(Records, metaclass=Singleton):
 
 
 def parse_package_name(name):
-    m = re.match("([^-]*)-(.*)\.([^\.]*)$", name)
+    m = re.match(r"([^-]*)-(.*)\.([^\.]*)$", name)
     if m:
         return m[1], m[2], m[3]
     else:
-        m = re.match("([^-]*)\.([^\.]*)$", name)
+        m = re.match(r"([^-]*)\.([^\.]*)$", name)
         if m:
             return m[1], "", m[2]
         else:
@@ -1048,7 +1078,7 @@ class Package(Record):
             }
             if not self._metadata["basename"] in self.plural.groups:
                 self.plural.groups[self._metadata["basename"]] = []
-            if not self in self.plural.groups[self._metadata["basename"]]:
+            if self not in self.plural.groups[self._metadata["basename"]]:
                 self.plural.groups[self._metadata["basename"]].append(self)
 
         return self._metadata
@@ -1056,55 +1086,63 @@ class Package(Record):
     def refresh_related(self):
         related = {}
         patchsoftwaretitles = jamf_records(PatchSoftwareTitles)
-        patchsoftwaretitles_ids = {}
+        patchsoftwaretitles_definitions = {}
         for title in patchsoftwaretitles:
             pkgs = title.get_path("versions/version/package/name")
             versions = title.get_path("versions/version/software_version")
             if pkgs:
                 for ii, pkg in enumerate(pkgs):
-                    if pkg == None:
+                    if pkg is None:
                         continue
-                    if not str(title.id) in patchsoftwaretitles_ids:
-                        patchsoftwaretitles_ids[str(title.id)] = {}
-                    patchsoftwaretitles_ids[str(title.id)][versions[ii]] = pkg
-                    if not pkg in related:
+                    if not str(title.id) in patchsoftwaretitles_definitions:
+                        patchsoftwaretitles_definitions[str(title.id)] = {}
+                    patchsoftwaretitles_definitions[str(title.id)][versions[ii]] = pkg
+                    if pkg not in related:
                         related[pkg] = {"PatchSoftwareTitles": []}
-                    if not "PatchSoftwareTitles" in related[pkg]:
+                    if "PatchSoftwareTitles" not in related[pkg]:
                         related[pkg]["PatchSoftwareTitles"] = []
                     temp = title.name + " - " + versions[ii]
                     related[pkg]["PatchSoftwareTitles"].append(temp)
         patchpolicies = jamf_records(PatchPolicies)
         for policy in patchpolicies:
-            parent_id = policy.get_path("software_title_configuration_id")[0]
-            parent_version = policy.get_path("general/target_version")[0]
-            if str(parent_id) in patchsoftwaretitles_ids:
-                ppp = patchsoftwaretitles_ids[str(parent_id)]
-                if parent_version in ppp:
-                    pkg = ppp[parent_version]
-                    if not pkg in related:
+            patchsoftwaretitle_id = policy.get_path("software_title_configuration_id")
+            parent_pkg_version = policy.get_path("general/target_version")
+            if str(patchsoftwaretitle_id) in patchsoftwaretitles_definitions:
+                patch_definitions = patchsoftwaretitles_definitions[
+                    str(patchsoftwaretitle_id)
+                ]
+                if parent_pkg_version in patch_definitions:
+                    pkg = patch_definitions[parent_pkg_version]
+                    if pkg not in related:
                         related[pkg] = {"PatchPolicies": []}
-                    if not "PatchPolicies" in related[pkg]:
+                    if "PatchPolicies" not in related[pkg]:
                         related[pkg]["PatchPolicies"] = []
                     related[pkg]["PatchPolicies"].append(policy.name)
         policies = jamf_records(Policies)
         for policy in policies:
-            pkgs = policy.get_path("package_configuration/packages/package/name")
+            try:
+                pkgs = policy.get_path("package_configuration/packages/package/name")
+            except NotFound:
+                pkgs = []
             if pkgs:
                 for pkg in pkgs:
-                    if not pkg in related:
+                    if pkg not in related:
                         related[pkg] = {"Policies": []}
-                    if not "Policies" in related[pkg]:
+                    if "Policies" not in related[pkg]:
                         related[pkg]["Policies"] = []
                     related[pkg]["Policies"].append(policy.name)
         groups = jamf_records(ComputerGroups)
         for group in groups:
-            criterions = group.get_path("criteria/criterion/value")
+            try:
+                criterions = group.get_path("criteria/criterion/value")
+            except NotFound:
+                criterions = None
             if criterions:
                 for pkg in criterions:
                     if pkg and re.search(".pkg|.zip|.dmg", pkg[-4:]):
-                        if not pkg in related:
+                        if pkg not in related:
                             related[pkg] = {"ComputerGroups": []}
-                        if not "ComputerGroups" in related[pkg]:
+                        if "ComputerGroups" not in related[pkg]:
                             related[pkg]["ComputerGroups"] = []
                         related[pkg]["ComputerGroups"].append(group.name)
         self.__class__._related = related
@@ -1203,41 +1241,44 @@ class PatchSoftwareTitle(Record):
         if not type(versions) is list:
             versions = [versions]
         for version in versions:
-            if version["package"] != None:
+            if version["package"] is not None:
                 print(f" {version['software_version']}: {version['package']['name']}")
 
     def patchpolicies_print_during(self):
         print(self.name)
         patchpolicies = jamf_records(PatchPolicies)
         for policy in patchpolicies:
-            parent_id = policy.get_path("software_title_configuration_id")[0]
-            if str(parent_id) != str(self.id):
+            try:
+                policy_id = policy.get_path("software_title_configuration_id")
+            except NotFound:
+                policy_id = None
+            if str(policy_id) != str(self.id):
                 continue
             print(f" {policy.data['general']['target_version']}: {str(policy)}")
 
     def set_all_packages_update_during(self):
         policy_regex = {
-            "1Password 7": "^1Password-%VERSION%\.pkg",
-            "Apple GarageBand 10": "^GarageBand-%VERSION%\.pkg",
-            "Apple Keynote": "^Keynote-%VERSION%\.pkg",
-            "Apple Numbers": "^Numbers-%VERSION%\.pkg",
-            "Apple Pages": "^Pages-%VERSION%\.pkg",
-            "Apple Xcode": "^Xcode-%VERSION%\.pkg",
-            "Apple iMovie": "^iMovie-%VERSION%\.pkg",
-            "Arduino IDE": "^Arduino-%VERSION%\.pkg",
-            "Bare Bones BBEdit": "BBEdit-%VERSION%\.pkg",
-            "BusyCal 3": "^BusyCal-%VERSION%\.pkg",
-            "Microsoft Remote Desktop 10": "^Microsoft Remote Desktop-%VERSION%\.pkg",
-            "Microsoft Visual Studio Code": "^Visual Studio Code-%VERSION%\.pkg",
-            "Microsoft Teams": "^Microsoft_Teams_%VERSION%\.pkg",
-            "Mozilla Firefox": "^Firefox-%VERSION%\.pkg",
-            "R for Statistical Computing": "^R-%VERSION%\.pkg",
-            "RStudio Desktop": "RStudio-%VERSION%\.dmg",
-            "Sublime Text 3": "Sublime Text-%VERSION%\.pkg",
-            "VLC media player": "VLC-%VERSION%\.pkg",
-            "VMware Fusion 12": "VMware Fusion-%VERSION%\.pkg",
-            "VMware Horizon 8 Client": "VMwareHorizonClient-%VERSION%.pkg",
-            "Zoom Client for Meetings": "Zoom-%VERSION%.pkg",
+            "1Password 7": r"^1Password-%VERSION%\.pkg",
+            "Apple GarageBand 10": r"^GarageBand-%VERSION%\.pkg",
+            "Apple Keynote": r"^Keynote-%VERSION%\.pkg",
+            "Apple Numbers": r"^Numbers-%VERSION%\.pkg",
+            "Apple Pages": r"^Pages-%VERSION%\.pkg",
+            "Apple Xcode": r"^Xcode-%VERSION%\.pkg",
+            "Apple iMovie": r"^iMovie-%VERSION%\.pkg",
+            "Arduino IDE": r"^Arduino-%VERSION%\.pkg",
+            "Bare Bones BBEdit": r"BBEdit-%VERSION%\.pkg",
+            "BusyCal 3": r"^BusyCal-%VERSION%\.pkg",
+            "Microsoft Remote Desktop 10": r"^Microsoft Remote Desktop-%VERSION%\.pkg",
+            "Microsoft Visual Studio Code": r"^Visual Studio Code-%VERSION%\.pkg",
+            "Microsoft Teams": r"^Microsoft_Teams_%VERSION%\.pkg",
+            "Mozilla Firefox": r"^Firefox-%VERSION%\.pkg",
+            "R for Statistical Computing": r"^R-%VERSION%\.pkg",
+            "RStudio Desktop": r"RStudio-%VERSION%\.dmg",
+            "Sublime Text 3": r"Sublime Text-%VERSION%\.pkg",
+            "VLC media player": r"VLC-%VERSION%\.pkg",
+            "VMware Fusion 12": r"VMware Fusion-%VERSION%\.pkg",
+            "VMware Horizon 8 Client": r"VMwareHorizonClient-%VERSION%.pkg",
+            "Zoom Client for Meetings": r"Zoom-%VERSION%.pkg",
         }
         change_made = False
         packages = jamf_records(Packages)
@@ -1253,7 +1294,9 @@ class PatchSoftwareTitle(Record):
                             "%VERSION%", pkg_version["software_version"]
                         )
                     else:
-                        regex = f".*{self.name}.*{pkg_version['software_version']}\.pkg"
+                        regex = (
+                            rf".*{self.name}.*{pkg_version['software_version']}\.pkg"
+                        )
                     regex = regex.replace("(", "\\(")
                     regex = regex.replace(")", "\\)")
                     if re.search(regex, package.name):
@@ -1280,7 +1323,7 @@ class PatchSoftwareTitle(Record):
         if not type(versions) is list:
             versions = [versions]
         for version in versions:
-            if version["package"] != None:
+            if version["package"] is not None:
                 print(f" {version['software_version']}: {version['package']['name']}")
             else:
                 print(f" {version['software_version']}: -")
@@ -1339,7 +1382,10 @@ class Policy(Record):
             _trigger.append("Startup")
         if self.data["general"]["trigger_login"] == "true":
             _trigger.append("Login")
-        if self.data["general"]["trigger_logout"] == "true":
+        if (
+            hasattr(self.data["general"], "trigger_logout")
+            and self.data["general"]["trigger_logout"] == "true"
+        ):
             _trigger.append("Logout")
         if self.data["general"]["trigger_network_state_changed"] == "true":
             _trigger.append("Network")
@@ -1424,13 +1470,10 @@ class Policy(Record):
         return _text
 
     def promote_update_during(self):
-        if not "package" in self.data["package_configuration"]["packages"]:
+        if "package" not in self.data["package_configuration"]["packages"]:
             print(f"{self.name} has no package to update.")
             return True
-        if int(self.data["package_configuration"]["packages"]["size"]) > 1:
-            my_packages = self.data["package_configuration"]["packages"]["package"]
-        else:
-            my_packages = [self.data["package_configuration"]["packages"]["package"]]
+        my_packages = self.data["package_configuration"]["packages"]["package"]
 
         all_packages = jamf_records(Packages)
         print(self.name)
@@ -1512,8 +1555,11 @@ class Script(Record):
     plural_class = "Scripts"
 
     def script_contents_print_during(self):
-        printme = self.get_path("script_contents")
-        print(printme[0])
+        try:
+            printme = self.get_path("script_contents")
+            print(printme[0])
+        except NotFound:
+            printme = None
 
 
 class Scripts(Records, metaclass=Singleton):
